"""Stormgate replay parsing tools"""

from __future__ import annotations

from dataclasses import dataclass, field
import gzip
from io import BytesIO
import logging
import struct
from collections import defaultdict
from contextlib import contextmanager
from enum import IntEnum
from pathlib import Path
from typing import BinaryIO, Dict, Iterable, List, Optional, Union
from uuid import UUID

from pydantic import BaseModel

from . import stormgate_pb2 as pb

logger = logging.getLogger(__name__)


# My empirical testing found that the timestamps in replays seem to be in units of ~0.976ms.
# This is very close to 1/1024 of a second, so we'll assume that this is the exact value.
REPLAY_TIMESTAMP_UNIT = 1 / 1024

FRIGATE = 55366


@contextmanager
def decompress(replay: Union[Path, BinaryIO]):
    """Open a gzipped stormgate replay, skipping the 16-byte header."""
    if isinstance(replay, Path):
        replay = replay.open("rb")
    with replay:
        replay.seek(16)
        yield BytesIO(gzip.decompress(replay.read()))


def get_build_number(replay: Union[Path, BinaryIO]) -> int:
    """Find the Stormgate version number that produced a given replay file.

    This is the number that can be found at the start of unrenamed replays,
    e.g. 44420 in CL44420-2024.01.31-16.23.SGReplay; but it is also stored in
    the 16-byte header, so we get it from there instead."""
    if isinstance(replay, Path):
        replay = replay.open("rb")
    replay.seek(12)
    (x,) = struct.unpack("<i", replay.read(4))
    replay.seek(0)
    return x


def read_varint(f: BinaryIO) -> Optional[int]:
    """Read a base-7 varint from a binary stream, or return None if the stream is at EOF."""
    # This recursive implementation would be slow for large varints; but since almost
    # all replay chunks are <128 bytes, we very rarely actually recurse; and on averaget
    # his ends up signficantly faster than an iterative implementation.
    bs = f.read(1)
    if len(bs) == 0:
        return None
    byte = ord(bs)
    digits = byte & 0b01111111
    if byte & 0b10000000:
        nxt = read_varint(f)
        if nxt is None:
            raise ValueError("EOF encountered while parsing varint")
        digits += nxt << 7
    return digits


def split_replay(replay: Union[Path, BinaryIO]) -> Iterable[bytes]:
    """Split a replay into a sequence of chunks, each of which is a raw
    bytestring containing a wire-format encoding of a protobuf message."""
    with decompress(replay) as f:
        while True:
            length = read_varint(f)
            if length is None:
                break
            yield f.read(length)


class Spectator(BaseModel):
    nickname: str
    nickname_discriminator: Optional[str] = None
    uuid: Optional[UUID] = None


class Player(BaseModel):
    nickname: str
    nickname_discriminator: Optional[str] = None
    uuid: Optional[UUID] = None
    faction: Optional[str] = None
    is_ai: bool = False
    disconnect_time: Optional[float] = None
    leave_reason: str = "unknown"


class ReplaySummary(BaseModel):
    build_number: int
    map_name: Optional[str]
    players: List[Player] = []
    spectators: List[Spectator] = []
    duration_seconds: Optional[float] = None
    is_1v1_ladder_game: bool = False


def summarize_replay(replay: Union[Path, BinaryIO]) -> ReplaySummary:
    """Parse what we can from a stormgate replay."""
    build_number = get_build_number(replay)
    state = GameState.at_end_of(replay)
    info = ReplaySummary(
        build_number=build_number,
        map_name=state.map_name,
    )
    if state.game_started_time is not None:
        first_left_game_time = min(
            c.left_game_time or float("inf") for c in state.clients.values()
        )
        info.duration_seconds = (
            first_left_game_time - state.game_started_time
        ) * REPLAY_TIMESTAMP_UNIT
    for slot in state.slots.values():
        if slot.ai_type is not None:
            info.players.append(
                Player(
                    nickname=slot.ai_type.name,
                    is_ai=True,
                    faction=slot.faction.name,
                )
            )
        elif slot.client_id is not None:
            client = state.clients.pop(slot.client_id)
            if client.nickname is None:
                raise ReplayParsingError(
                    f"No nickname found for client {client.client_id} = {client.uuid}?"
                )
            info.players.append(
                p := Player(
                    nickname=client.nickname,
                    nickname_discriminator=client.discriminator,
                    uuid=client.uuid,
                    is_ai=False,
                    faction=slot.faction.name,
                )
            )
            if (
                state.game_started_time is not None
                and client.left_game_time is not None
            ):
                p.disconnect_time = (
                    client.left_game_time - state.game_started_time
                ) * REPLAY_TIMESTAMP_UNIT
            p.leave_reason = client.left_game_reason.name
    for client in state.clients.values():
        if client.slot_number != 255:
            raise ReplayParsingError("Player not in a slot but slot_number != 255?")
        if client.nickname is None:
            raise ReplayParsingError(
                f"No nickname found for client {client.client_id} = {client.uuid}?"
            )
        info.spectators.append(
            Spectator(
                nickname=client.nickname,
                nickname_discriminator=client.discriminator,
                uuid=client.uuid,
            )
        )
    info.is_1v1_ladder_game = (  # TODO: Can we update this with the match type?
        len(info.players) == 2
        and len(info.spectators) == 0
        and len(state.slot_assignments) > 0
    )
    return info


# Unfortunately, to correctly determine who's in player slots and who's in
# spectator slots we need to know how many players the map has; and this
# metadata is not included in the replay.
# For now, we just hardcode the known co-op maps.
player_slot_count: Dict[str, int] = defaultdict(
    lambda: 2,
    {
        "WreckHavoc": 3,
        "TheAbyssalGates": 3,
        "TurfWar": 3,
        "IsleOfDread": 3,
        "RitualWoods": 3,
        "InfestedCrater": 3,
        "CrookedCanyon": 3,
    },
)


class SlotType(IntEnum):
    Closed = 0
    Human = 1
    Ai = 2


class Faction(IntEnum):
    Vanguard = 0
    Infernals = 1
    Celestial = 2
    Blockade = 101
    Amara = 102
    Maloc = 201
    Warz = 202
    Auralanna = 301


class AIType(IntEnum):
    PeacefulBot = 0
    MurderBotJr = 1
    MurderBotSr = 2


@dataclass()
class Slot:
    type: SlotType = SlotType.Human
    faction: Faction = Faction(0)
    ai_type: Optional[AIType] = None
    client_id: Optional[int] = None


# The python protobuf bindings don't use standard python enums - they just return ints.
# I want nice type hints so I'm just going to maintain this manually.
class LeftGameReason(IntEnum):
    Unknown = 0
    Surrender = 1
    Leave = 2
    Disconnect = 3


class MatchType(IntEnum):
    Unknown = 0
    Custom = 1
    Ranked1v1 = 2
    Coop3ve = 3


@dataclass()
class Client:
    uuid: UUID
    client_id: int
    nickname: Optional[str] = None
    discriminator: Optional[str] = None
    slot_number: Optional[int] = None  # 255 means spectator
    left_game_time: Optional[float] = None
    left_game_reason: LeftGameReason = LeftGameReason.Unknown


@dataclass()
class SlotAssignment:
    slot_number: int
    nickname: str


def parse_uuid(uuid: pb.UUID) -> UUID:
    return UUID(bytes=struct.pack(">QQ", uuid.part1, uuid.part2))


@dataclass()
class GameState:
    """Stormgate match state machine - reads commands from replay and updates state"""

    map_name: Optional[str] = None
    match_type: MatchType = MatchType.Unknown
    slots: Dict[int, Slot] = field(default_factory=dict)
    clients: Dict[int, Client] = field(default_factory=dict)
    slot_assignments: Dict[UUID, SlotAssignment] = field(default_factory=dict)
    game_started: bool = False
    game_started_time: Optional[float] = None

    @classmethod
    def at_end_of(cls, replay: Union[Path, BinaryIO]) -> GameState:
        """Simulate an entire replay and return the end state."""
        self = cls()
        # Using split_replay adds a significant overhead to the runtime of this
        # function; so I've inlined this loop instead.
        with decompress(replay) as f:
            while length := read_varint(f):
                self.process(pb.ReplayChunk.FromString(f.read(length)))
        return self

    def process(self, chunk: pb.ReplayChunk):
        """Update the state using a single replay chunk/command."""
        content = chunk.inner.content
        content_type = content.WhichOneof("content_type")
        if not content_type:
            return
        msg = getattr(content, content_type)
        client_id = chunk.client_id
        timestamp = chunk.timestamp

        if isinstance(msg, pb.MapDetails):
            self.map_name = msg.map_name
            self.match_type = MatchType(msg.match_type)

            if self.match_type == MatchType.Ranked1v1:
                slot_count = 2
            elif self.match_type == MatchType.Coop3ve:
                slot_count = 3
            else:
                slot_count = player_slot_count[msg.map_name]  # Fall back to old sysyem

            logger.debug(f"MatchType: {self.match_type}")

            logger.debug(f"Setting up {slot_count} slots for map {self.map_name}")
            for i in range(1, slot_count + 1):
                self.slots[i] = Slot()

        elif isinstance(msg, pb.AssignPlayerSlot):
            self.slot_assignments[parse_uuid(msg.uuid)] = SlotAssignment(
                slot_number=msg.slot, nickname=msg.nickname
            )
            logger.debug(f"Assigning slot {msg.slot} to {msg.uuid}")

        elif isinstance(msg, pb.Player):
            self.clients[client_id] = client = Client(
                client_id=client_id,
                uuid=parse_uuid(msg.uuid),
                nickname=msg.name.nickname,
                discriminator=msg.name.discriminator,
            )
            logger.debug(
                f"Setting up player {client_id}: {client.nickname} {client.uuid}"
            )
            # If we're in a matchmaking game, the server has pre-assigned a slot for the player:
            if (assignment := self.slot_assignments.get(client.uuid)) is not None:
                client.slot_number = assignment.slot_number
                self.slots[assignment.slot_number].client_id = client_id
                logger.debug(
                    f"Putting player {client_id} in pre-assigned slot {client.slot_number}"
                )

        elif isinstance(msg, pb.ClientConnected):
            if not msg.HasField("uuid"):
                # We should have already filled in the player details in handle_player,
                # don't worry about it
                return
            self.clients[msg.client_id] = client = Client(
                client_id=msg.client_id,
                uuid=parse_uuid(msg.uuid),
            )
            logger.debug(f"Setting up player {client.client_id}: {client.uuid}")
            if (assignment := self.slot_assignments.get(client.uuid)) is not None:
                client.slot_number = assignment.slot_number
                if client.nickname is not None:
                    assert client.nickname == assignment.nickname
                else:
                    client.nickname = assignment.nickname
                self.slots[assignment.slot_number].client_id = client.client_id
                logger.debug(
                    f"Putting player {client.client_id} in pre-assigned slot {client.slot_number}"
                )

        elif isinstance(msg, pb.PlayerLeftGame):
            if self.game_started:
                self.clients[client_id].left_game_time = timestamp
                self.clients[client_id].left_game_reason = LeftGameReason(msg.reason)
            else:
                client = self.clients.pop(client_id, None)
                # In aborted ladder games, we sometimes get a left game before the player joined message
                suffix = (
                    f": {client.nickname} {client.uuid}" if client is not None else ""
                )
                logger.debug(f"Removing player {client_id}: {suffix}")
                for slot in self.slots.values():
                    if slot.client_id == client_id:
                        slot.client_id = None

        elif isinstance(msg, pb.ClientDisconnected):
            if self.game_started:
                client = self.clients[msg.client_id]
                if client.left_game_time is None:
                    # Server telling us client disconnected without leaving cleanly
                    assert client.uuid == parse_uuid(msg.player_uuid)
                    client.left_game_time = timestamp
                    client.left_game_reason = LeftGameReason(msg.reason)

        elif isinstance(msg, pb.LobbyChangeSlot):
            if not self.slots:
                raise ReplayParsingError("Received slot change before map info?")
            client = self.clients[client_id]
            if client.slot_number is not None and client.slot_number != 255:
                self.slots[client.slot_number].client_id = None
            if msg.choice.WhichOneof("choice_type") == "specific_slot":
                slot_number = msg.choice.specific_slot.slot
            else:
                # Client did not choose a specific slot, so they get the first open human slot
                for slot_number, slot in self.slots.items():
                    if slot.type == SlotType.Human and slot.client_id is None:
                        break
                else:
                    # No open slots, become spectator
                    slot_number = 255
            client.slot_number = slot_number
            if slot_number != 255:
                slot = self.slots[slot_number]
                if slot.type != SlotType.Human:
                    raise ReplayParsingError("Client assigned to non-human slot?")
                if slot.client_id is not None:
                    raise ReplayParsingError(
                        "Client assigned to already occupied slot?"
                    )
                slot.client_id = client_id
                logger.debug(f"Putting player {client_id} in slot {slot_number}")

        elif isinstance(msg, pb.LobbySetVariable):
            slot = self.slots[msg.slot]
            key = msg.variable_id
            value = msg.value
            if key == 374945738:
                slot.type = SlotType(value)
                logger.debug(f"Set slot[{msg.slot}].type = {slot.type}")
                if slot.type == SlotType.Ai:
                    slot.ai_type = AIType(0)
                    logger.debug(f"Set slot[{msg.slot}].ai_type = {slot.ai_type}")
                else:
                    slot.ai_type = None
                    logger.debug(f"Set slot[{msg.slot}].ai_type = None")
            elif key == 2952722564:
                slot.faction = Faction(value)
                logger.debug(f"Set slot[{msg.slot}].faction = {slot.faction}")
            elif key == 655515685:
                slot.ai_type = AIType(value)
                logger.debug(f"Set slot[{msg.slot}].ai_type = {slot.ai_type}")

        elif isinstance(msg, pb.StartGame):
            logger.debug("Marking game as started")
            self.game_started = True
            self.game_started_time = float(timestamp)


class ReplayParsingError(Exception):
    pass



import logging

from rich.console import Console
from rich.logging import RichHandler

_logging_configured: bool = False


def configure_logging(debug: bool = False):
    global _logging_configured
    if not _logging_configured:
        _logging_configured = True
        level = logging.DEBUG if debug else logging.INFO
        console = Console(stderr=True)
        logging.captureWarnings(True)
        logging.basicConfig(
            level=level,
            format="%(name)s: %(message)s" if debug else "%(message)s",
            handlers=[RichHandler(console=console, show_path=False, show_time=False)],
        )



from pathlib import Path
from typing import NamedTuple

import pytest

data_dir = Path(__file__).parent / "replays"


def pytest_addoption(parser):
    parser.addoption("--update-golden", default=False, action="store_true")


class ReplayCase(NamedTuple):
    replay_file: Path
    summary_file: Path
    expected_name_file: Path


@pytest.fixture(params=list(data_dir.glob("**/*.SGReplay")), ids=lambda c: c.stem)
def replay_case(request):
    path = request.param
    yield ReplayCase(
        replay_file=path,
        summary_file=path.with_suffix(".json"),
        expected_name_file=path.with_suffix(".txt"),
    )



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

Coop3vE: MatchType
Custom: MatchType
DESCRIPTOR: _descriptor.FileDescriptor
Disconnect: LeaveReason
Leave: LeaveReason
Ranked1v1: MatchType
Surrender: LeaveReason
UnknownMatchType: MatchType
UnknownReason: LeaveReason

class AssignPlayerSlot(_message.Message):
    __slots__ = ["nickname", "slot", "uuid"]
    NICKNAME_FIELD_NUMBER: _ClassVar[int]
    SLOT_FIELD_NUMBER: _ClassVar[int]
    UUID_FIELD_NUMBER: _ClassVar[int]
    nickname: str
    slot: int
    uuid: UUID
    def __init__(self, uuid: _Optional[_Union[UUID, _Mapping]] = ..., slot: _Optional[int] = ..., nickname: _Optional[str] = ...) -> None: ...

class ClientConnected(_message.Message):
    __slots__ = ["client_id", "uuid"]
    CLIENT_ID_FIELD_NUMBER: _ClassVar[int]
    UUID_FIELD_NUMBER: _ClassVar[int]
    client_id: int
    uuid: UUID
    def __init__(self, client_id: _Optional[int] = ..., uuid: _Optional[_Union[UUID, _Mapping]] = ...) -> None: ...

class ClientDisconnected(_message.Message):
    __slots__ = ["client_id", "player_uuid", "reason"]
    CLIENT_ID_FIELD_NUMBER: _ClassVar[int]
    PLAYER_UUID_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    client_id: int
    player_uuid: UUID
    reason: LeaveReason
    def __init__(self, client_id: _Optional[int] = ..., reason: _Optional[_Union[LeaveReason, str]] = ..., player_uuid: _Optional[_Union[UUID, _Mapping]] = ...) -> None: ...

class LobbyChangeSlot(_message.Message):
    __slots__ = ["choice"]
    class SlotChoice(_message.Message):
        __slots__ = ["specific_slot"]
        class SpecificSlot(_message.Message):
            __slots__ = ["slot"]
            SLOT_FIELD_NUMBER: _ClassVar[int]
            slot: int
            def __init__(self, slot: _Optional[int] = ...) -> None: ...
        SPECIFIC_SLOT_FIELD_NUMBER: _ClassVar[int]
        specific_slot: LobbyChangeSlot.SlotChoice.SpecificSlot
        def __init__(self, specific_slot: _Optional[_Union[LobbyChangeSlot.SlotChoice.SpecificSlot, _Mapping]] = ...) -> None: ...
    CHOICE_FIELD_NUMBER: _ClassVar[int]
    choice: LobbyChangeSlot.SlotChoice
    def __init__(self, choice: _Optional[_Union[LobbyChangeSlot.SlotChoice, _Mapping]] = ...) -> None: ...

class LobbySetVariable(_message.Message):
    __slots__ = ["slot", "value", "variable_id"]
    SLOT_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    VARIABLE_ID_FIELD_NUMBER: _ClassVar[int]
    slot: int
    value: int
    variable_id: int
    def __init__(self, slot: _Optional[int] = ..., variable_id: _Optional[int] = ..., value: _Optional[int] = ...) -> None: ...

class MapDetails(_message.Message):
    __slots__ = ["map_folder", "map_name", "map_seed", "match_type"]
    MAP_FOLDER_FIELD_NUMBER: _ClassVar[int]
    MAP_NAME_FIELD_NUMBER: _ClassVar[int]
    MAP_SEED_FIELD_NUMBER: _ClassVar[int]
    MATCH_TYPE_FIELD_NUMBER: _ClassVar[int]
    map_folder: str
    map_name: str
    map_seed: int
    match_type: MatchType
    def __init__(self, map_folder: _Optional[str] = ..., map_name: _Optional[str] = ..., map_seed: _Optional[int] = ..., match_type: _Optional[_Union[MatchType, str]] = ...) -> None: ...

class Player(_message.Message):
    __slots__ = ["name", "uuid"]
    class PlayerName(_message.Message):
        __slots__ = ["discriminator", "nickname"]
        DISCRIMINATOR_FIELD_NUMBER: _ClassVar[int]
        NICKNAME_FIELD_NUMBER: _ClassVar[int]
        discriminator: str
        nickname: str
        def __init__(self, nickname: _Optional[str] = ..., discriminator: _Optional[str] = ...) -> None: ...
    NAME_FIELD_NUMBER: _ClassVar[int]
    UUID_FIELD_NUMBER: _ClassVar[int]
    name: Player.PlayerName
    uuid: UUID
    def __init__(self, uuid: _Optional[_Union[UUID, _Mapping]] = ..., name: _Optional[_Union[Player.PlayerName, _Mapping]] = ...) -> None: ...

class PlayerLeftGame(_message.Message):
    __slots__ = ["player_uuid", "reason"]
    PLAYER_UUID_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    player_uuid: UUID
    reason: LeaveReason
    def __init__(self, player_uuid: _Optional[_Union[UUID, _Mapping]] = ..., reason: _Optional[_Union[LeaveReason, str]] = ...) -> None: ...

class ReplayChunk(_message.Message):
    __slots__ = ["client_id", "inner", "timestamp"]
    class Wrapper(_message.Message):
        __slots__ = ["content"]
        class ReplayContent(_message.Message):
            __slots__ = ["assign_player_slot", "change_slot", "client_connected", "client_disconnected", "map_details", "player", "player_left_game", "set_variable", "start_game"]
            ASSIGN_PLAYER_SLOT_FIELD_NUMBER: _ClassVar[int]
            CHANGE_SLOT_FIELD_NUMBER: _ClassVar[int]
            CLIENT_CONNECTED_FIELD_NUMBER: _ClassVar[int]
            CLIENT_DISCONNECTED_FIELD_NUMBER: _ClassVar[int]
            MAP_DETAILS_FIELD_NUMBER: _ClassVar[int]
            PLAYER_FIELD_NUMBER: _ClassVar[int]
            PLAYER_LEFT_GAME_FIELD_NUMBER: _ClassVar[int]
            SET_VARIABLE_FIELD_NUMBER: _ClassVar[int]
            START_GAME_FIELD_NUMBER: _ClassVar[int]
            assign_player_slot: AssignPlayerSlot
            change_slot: LobbyChangeSlot
            client_connected: ClientConnected
            client_disconnected: ClientDisconnected
            map_details: MapDetails
            player: Player
            player_left_game: PlayerLeftGame
            set_variable: LobbySetVariable
            start_game: StartGame
            def __init__(self, map_details: _Optional[_Union[MapDetails, _Mapping]] = ..., client_connected: _Optional[_Union[ClientConnected, _Mapping]] = ..., player: _Optional[_Union[Player, _Mapping]] = ..., change_slot: _Optional[_Union[LobbyChangeSlot, _Mapping]] = ..., set_variable: _Optional[_Union[LobbySetVariable, _Mapping]] = ..., start_game: _Optional[_Union[StartGame, _Mapping]] = ..., player_left_game: _Optional[_Union[PlayerLeftGame, _Mapping]] = ..., client_disconnected: _Optional[_Union[ClientDisconnected, _Mapping]] = ..., assign_player_slot: _Optional[_Union[AssignPlayerSlot, _Mapping]] = ...) -> None: ...
        CONTENT_FIELD_NUMBER: _ClassVar[int]
        content: ReplayChunk.Wrapper.ReplayContent
        def __init__(self, content: _Optional[_Union[ReplayChunk.Wrapper.ReplayContent, _Mapping]] = ...) -> None: ...
    CLIENT_ID_FIELD_NUMBER: _ClassVar[int]
    INNER_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    client_id: int
    inner: ReplayChunk.Wrapper
    timestamp: int
    def __init__(self, timestamp: _Optional[int] = ..., client_id: _Optional[int] = ..., inner: _Optional[_Union[ReplayChunk.Wrapper, _Mapping]] = ...) -> None: ...

class StartGame(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class UUID(_message.Message):
    __slots__ = ["part1", "part2"]
    PART1_FIELD_NUMBER: _ClassVar[int]
    PART2_FIELD_NUMBER: _ClassVar[int]
    part1: int
    part2: int
    def __init__(self, part1: _Optional[int] = ..., part2: _Optional[int] = ...) -> None: ...

class MatchType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class LeaveReason(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



"""Rename stormgate replays to include useful info in filename"""

from __future__ import annotations

import logging
import os
import platform
import re
import string
from collections import defaultdict
from datetime import datetime, timedelta, timezone, tzinfo
from pathlib import Path
from shutil import copytree, rmtree
from typing import Iterable, NamedTuple, Optional, Union
from uuid import UUID

from packaging import version
from typing_extensions import Literal

from shroudstone import __version__
from shroudstone.config import data_dir
from shroudstone.replay import FRIGATE, Player, ReplaySummary, summarize_replay

logger = logging.getLogger(__name__)


TOLERANCE = timedelta(seconds=90)
"""Maximum time difference to consider games a match"""

BAD_CHARS = re.compile(r'[<>:"/\\|?*\0]')
"""Characters forbidden in filenames on Linux or Windows"""

skipped_replays_file = data_dir / "skipped_replays.txt"
"""Directory in which previouslyskipped replays are recorded"""


VALID_FIELDS = {
    "1v1": [
        "us",
        "them",
        "f1",
        "f2",
        "r1",  # for backwards compat
        "r2",
        "time",
        "duration",
        "result",
        "map_name",
        "build_number",
    ],
    "generic": [
        "players",
        "players_with_factions",
        "time",
        "duration",
        "map_name",
        "build_number",
    ],
}


def migrate():
    last_run_version_file = data_dir / "last_run_version.txt"
    if last_run_version_file.exists():
        last_run_version = version.parse(
            last_run_version_file.read_text(encoding="utf-8")
        )
    else:
        last_run_version = version.parse("0.1.0a29")
    if last_run_version < version.parse("0.1.0a30"):
        logger.info(
            "Cache files are from an incompatible version of shroudstone, deleting them."
        )
        skipped_replays_file.unlink(missing_ok=True)
        rmtree(data_dir / "stormgateworld-cache", ignore_errors=True)
    last_run_version_file.write_text(__version__, encoding="utf-8")


def rename_replays(
    replay_dir: Path,
    format_1v1: str,
    format_generic: str,
    dry_run: bool = False,
    backup: bool = True,
    reprocess: bool = False,
    files: Optional[Iterable[Path]] = None,
):
    migrate()
    if dry_run:
        # Don't bother
        bu_dir = None
        logger.warning(
            "Performing a dry run - will show what would happen but not actually touch anything."
        )
    elif backup:
        bu_dir = replay_dir.parent / f"{replay_dir.name}.backup"
        backup_dir(replay_dir, bu_dir)
    else:
        bu_dir = None
        logger.warning("No backup being made! You asked for it!")

    if files is None:
        if reprocess:
            # Reprocess all replays
            pattern = "**/*.SGReplay"
            logger.info(f"Searching for all replays in {replay_dir}.")
        else:
            # Only look for replays we haven't already renamed
            pattern = "**/CL*.SGReplay"
            logger.info(f"Searching for unrenamed replays in {replay_dir}.")

        files = replay_dir.glob(pattern)

    def try_parse(path: Path):
        try:
            return Replay.from_path(path)
        except Exception:
            logger.exception(f"Unexpected error parsing {path}")

    replays = [x for x in map(try_parse, files) if x is not None]
    if not replays:
        logger.warning(
            "No new replays found to rename! "
            f"If you weren't expecting this, check your replay_dir '{replay_dir}' is correct."
        )
        return

    n = len(replays)
    earliest_time = min(x.time for x in replays)
    logger.info(f"Found {n} unrenamed replays going back to {earliest_time}.")
    if not skipped_replays_file.exists():
        skipped_replays_file.touch()
    previously_skipped_paths = {
        Path(x)
        for x in skipped_replays_file.read_text(encoding="utf-8").splitlines()
        if x
    }
    skipped_paths = []

    counts = defaultdict(lambda: 0)
    for replay in replays:
        if replay.path in previously_skipped_paths:
            counts["skipped_old"] += 1
            logger.debug(
                f"We've previously skipped {replay.path.name}, so not commenting on it this time."
            )
        elif any(p.is_ai for p in replay.summary.players):
            counts["skipped_new"] += 1
            skipped_paths.append(replay.path)
            logger.info(f"{replay.path.name} is a game vs AI, skipping it.")
            continue
        else:
            try:
                newname = new_name_for(
                    replay, format_1v1=format_1v1, format_generic=format_generic
                )
                target = replay.path.parent / newname
                do_rename(replay.path, target, dry_run=dry_run)
                counts["renamed"] += 1
            except Exception as e:
                logger.error(f"Unexpected error handling {replay.path}: {e}")
                counts["error"] += 1

    if not dry_run:
        with skipped_replays_file.open("at", encoding="utf-8") as f:
            for path in skipped_paths:
                print(path, file=f)

    prefix = "DRY RUN: " if dry_run else ""
    counts_str = (
        "{renamed} replays renamed, "
        "{skipped_new} skipped (no matching ladder game), "
        "{skipped_ongoing} skipped (ongoing), "
        "{skipped_old} ignored (previously skipped), "
        "{error} errors."
    ).format_map(counts)
    logger.info(prefix + counts_str)


def backup_dir(replay_dir: Path, bu_dir: Path):
    logger.info(f"Backing up your replays to {bu_dir}.")
    copytree(replay_dir, bu_dir, dirs_exist_ok=True)


def naive_localtime_to_utc(dt: datetime) -> datetime:
    assert dt.tzinfo is None
    return dt.astimezone(timezone.utc).replace(tzinfo=None)


class Replay(NamedTuple):
    path: Path
    summary: ReplaySummary
    time: datetime
    us: Optional[Player]
    them: Optional[Player]

    @staticmethod
    def from_path(path: Path, filename_timezone: Optional[tzinfo] = None):
        # Original names use local times:
        if m := re.search(r"(\d\d\d\d)\.(\d\d)\.(\d\d)-(\d\d).(\d\d)", path.name):
            if filename_timezone is None:
                # Assume local time
                time = naive_localtime_to_utc(
                    datetime(*(int(x) for x in m.groups()))  # type: ignore
                )
            else:
                time = datetime(
                    *(int(x) for x in m.groups()), tzinfo=filename_timezone
                ).astimezone(timezone.utc)

        # Our renamed versions use UTC:
        elif m := re.search(r"(\d\d\d\d)-(\d\d)-(\d\d) (\d\d).(\d\d)", path.name):
            time = datetime(*(int(x) for x in m.groups()))  # type: ignore
        else:
            return None

        summary = summarize_replay(path)
        our_uuid = find_our_uuid(path)

        us = None
        them = None
        if len(summary.players) == 2:
            if summary.players[0].uuid == our_uuid:
                us, them = summary.players
            elif summary.players[1].uuid == our_uuid:
                them, us = summary.players
        elif len(summary.players) > 0 and summary.players[0].uuid == our_uuid:
            us = summary.players[0]

        return Replay(path=path, time=time, us=us, them=them, summary=summary)


def find_our_uuid(replay_path: Path) -> Optional[UUID]:
    """Given the path to a Stormgate replay, extract the player UUID. (This
    assumes it's stored in the usual directory heirarchy.)"""
    for part in reversed(replay_path.parts):
        try:
            return UUID(hex=part)
        except ValueError:
            pass


def get_result(replay: Replay):
    if not (replay.us and replay.them):
        return None
    if replay.summary.build_number >= FRIGATE:
        # Since Frigate we've had explicit surrender messages, so we rely on them alone for certainty:
        if replay.us.leave_reason == "surrender":
            return "loss"
        if replay.them.leave_reason == "surrender":
            return "win"
        return None
    else:
        # For old replays, best we can do is guess based on disconnection times:
        t1 = replay.us.disconnect_time
        t2 = replay.them.disconnect_time
        if t1 and t2:
            return "win" if t1 > t2 else "loss"
        elif t1:
            return "loss"
        elif t2:
            return "win"


def new_name_for(replay: Replay, format_1v1: str, format_generic: str) -> str:
    parts = {}
    parts["map_name"] = replay.summary.map_name
    parts["build_number"] = replay.summary.build_number
    duration = replay.summary.duration_seconds
    if duration is not None:
        minutes, seconds = divmod(int(duration), 60)
        parts["duration"] = f"{minutes:02d}m{seconds:02d}s"
    else:
        parts["duration"] = ""

    parts["time"] = replay.time

    us = replay.us
    them = replay.them
    if us and them:
        # 1v1
        parts["us"] = parts["p1"] = us.nickname
        parts["them"] = parts["p2"] = them.nickname

        parts["r1"] = parts["f1"] = (us.faction or "").capitalize()
        parts["r2"] = parts["f2"] = (them.faction or "").capitalize()

        result = get_result(replay)
        parts["result"] = (result or "unknown").capitalize()

        newname = format_1v1.format(**parts)
    else:
        parts["players"] = ", ".join(
            p.nickname.capitalize() for p in replay.summary.players
        )
        parts["players_with_factions"] = ", ".join(
            f"{p.nickname.capitalize()} {(p.faction or '').upper():.1}"
            for p in replay.summary.players
        )
        newname = format_generic.format(**parts)

    # In case we left some blanks, collapse multiple spaces to one space
    newname = re.sub(r"\s+", " ", newname)

    return sanitize_filename(newname)


def do_rename(source: Path, target: Path, dry_run: bool):
    if source == target:
        logger.debug(f"{source} already has the desired format, doing nothing :)")
        return

    if target.exists():
        logger.error(f"Not renaming {source}! {target} already exists!")
        return

    if dry_run:
        logger.info(f"DRY RUN: Would have renamed {source.name} => {target.name}.")
        return

    logger.info(f"Renaming {source.name} => {target.name}.")
    try:
        source.rename(target)
    except Exception as e:
        logger.error(f"Error renaming {source} => {target.name}: {e}")


def sanitize_filename(filename: str) -> str:
    """Remove bad characters from a filename"""
    return BAD_CHARS.sub("", filename)


def guess_replay_dir() -> Optional[Path]:
    if platform.system() == "Windows":
        # Should be easy, just look in the current user's local app data
        appdata = os.environ["LOCALAPPDATA"]
        paths = [Path(appdata) / "Stormgate" / "Saved" / "Replays"]
    else:
        # If this script is running on Linux and Stormgate is installed using
        # Steam+Proton, look in the steam compatdata:
        steammnt = (
            Path.home() / ".steam/root/steamapps/compatdata/2012510/pfx/dosdevices"
        )

        # If this script is running on the Windows Subsystem for Linux, we can
        # find the Windows drives mounted in /mnt:
        wslmnt = Path("/mnt")

        tail = "AppData/Local/Stormgate/Saved/Replays"
        paths = [
            *steammnt.glob(f"*/users/steamuser/{tail}"),
            *wslmnt.glob(f"*/Users/*/{tail}"),
            *wslmnt.glob(f"*/Documents and Settings/*/{tail}"),
        ]

    for path in paths:
        if path.is_dir():
            return path


def validate_format_string(
    format: str, type: Union[Literal["1v1"], Literal["generic"]]
):
    parts = string.Formatter().parse(format)
    for _, field_name, _, _ in parts:
        if field_name is not None and field_name not in VALID_FIELDS[type]:
            raise ValueError(f"Unknown replay field {field_name}")



import json

from shroudstone.replay import summarize_replay
from tests.conftest import ReplayCase

def test_replay_renaming(replay_case: ReplayCase, request):
    summary = summarize_replay(replay_case.replay_file).model_dump(mode="json")
    if request.config.getoption("--update-golden"):
        with replay_case.summary_file.open("w", encoding="utf-8") as f:
            json.dump(summary, f)
    else:
        with replay_case.summary_file.open("r", encoding="utf-8") as f:
            expected = json.load(f)
        assert summary == expected



from zoneinfo import ZoneInfo
from shroudstone.renamer import Replay, new_name_for
from tests.conftest import ReplayCase

f1v1 = "{time:%Y-%m-%d %H.%M} {result:.1} {duration} {us} {f1:.1}v{f2:.1} {them} - {map_name}.SGReplay"
fgeneric = (
    "{time:%Y-%m-%d %H.%M} {duration} {players_with_factions} - {map_name}.SGReplay"
)


def test_replay_renaming(replay_case: ReplayCase, request):
    # Since we have to resort to getting replay timestamps from the *local time*
    # in the filename, shroudstone.renamer's default behaviour is dependent upon the system timezone.
    # Thus for the purposes of a simple test suite we fix the timezone.
    replay = Replay.from_path(
        replay_case.replay_file, filename_timezone=ZoneInfo("Australia/Melbourne")
    )
    assert replay is not None
    new_name = new_name_for(replay, format_1v1=f1v1, format_generic=fgeneric)
    assert isinstance(new_name, str)
    if request.config.getoption("--update-golden"):
        replay_case.expected_name_file.write_text(new_name, encoding="utf-8")
    else:
        assert new_name == replay_case.expected_name_file.read_text(encoding="utf-8")



from tkinter.font import families, nametofont


def first_available_font(*names) -> str:
    fonts = families()
    for name in names:
        if name in fonts:
            return name
    return "times"


def setup_style(root):
    sans = first_available_font(
        "Ubuntu", "DejaVu Sans", "Sans", "Segoe UI", "Helvetica"
    )
    mono = first_available_font(
        "Iosevka",
        "DejaVu Sans Mono",
        "Ubuntu Mono",
        "Monaco",
        "Consolas",
        "Monospace",
    )
    nametofont("TkDefaultFont").configure(family=sans)
    nametofont("TkFixedFont").configure(family=mono)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: stormgate.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fstormgate.proto\x12\tstormgate\"\x9e\x05\n\x0bReplayChunk\x12\x11\n\ttimestamp\x18\x01 \x01(\x05\x12\x11\n\tclient_id\x18\x02 \x01(\x05\x12-\n\x05inner\x18\x03 \x01(\x0b\x32\x1e.stormgate.ReplayChunk.Wrapper\x1a\xb9\x04\n\x07Wrapper\x12=\n\x07\x63ontent\x18\x01 \x01(\x0b\x32,.stormgate.ReplayChunk.Wrapper.ReplayContent\x1a\xee\x03\n\rReplayContent\x12,\n\x0bmap_details\x18\x03 \x01(\x0b\x32\x15.stormgate.MapDetailsH\x00\x12\x36\n\x10\x63lient_connected\x18\x04 \x01(\x0b\x32\x1a.stormgate.ClientConnectedH\x00\x12#\n\x06player\x18\x0c \x01(\x0b\x32\x11.stormgate.PlayerH\x00\x12\x31\n\x0b\x63hange_slot\x18\r \x01(\x0b\x32\x1a.stormgate.LobbyChangeSlotH\x00\x12\x33\n\x0cset_variable\x18\x0f \x01(\x0b\x32\x1b.stormgate.LobbySetVariableH\x00\x12*\n\nstart_game\x18\x12 \x01(\x0b\x32\x14.stormgate.StartGameH\x00\x12\x35\n\x10player_left_game\x18\x19 \x01(\x0b\x32\x19.stormgate.PlayerLeftGameH\x00\x12<\n\x13\x63lient_disconnected\x18\x1f \x01(\x0b\x32\x1d.stormgate.ClientDisconnectedH\x00\x12\x39\n\x12\x61ssign_player_slot\x18% \x01(\x0b\x32\x1b.stormgate.AssignPlayerSlotH\x00\x42\x0e\n\x0c\x63ontent_type\"n\n\nMapDetails\x12\x12\n\nmap_folder\x18\x01 \x01(\t\x12\x10\n\x08map_name\x18\x02 \x01(\t\x12\x10\n\x08map_seed\x18\x03 \x01(\x05\x12(\n\nmatch_type\x18\x07 \x01(\x0e\x32\x14.stormgate.MatchType\"C\n\x0f\x43lientConnected\x12\x11\n\tclient_id\x18\x01 \x01(\x05\x12\x1d\n\x04uuid\x18\x02 \x01(\x0b\x32\x0f.stormgate.UUID\"\x8a\x01\n\x06Player\x12\x1d\n\x04uuid\x18\x02 \x01(\x0b\x32\x0f.stormgate.UUID\x12*\n\x04name\x18\x03 \x01(\x0b\x32\x1c.stormgate.Player.PlayerName\x1a\x35\n\nPlayerName\x12\x10\n\x08nickname\x18\x01 \x01(\t\x12\x15\n\rdiscriminator\x18\x02 \x01(\t\"\xd1\x01\n\x0fLobbyChangeSlot\x12\x35\n\x06\x63hoice\x18\x01 \x01(\x0b\x32%.stormgate.LobbyChangeSlot.SlotChoice\x1a\x86\x01\n\nSlotChoice\x12K\n\rspecific_slot\x18\x02 \x01(\x0b\x32\x32.stormgate.LobbyChangeSlot.SlotChoice.SpecificSlotH\x00\x1a\x1c\n\x0cSpecificSlot\x12\x0c\n\x04slot\x18\x01 \x01(\x05\x42\r\n\x0b\x63hoice_type\"D\n\x10LobbySetVariable\x12\x0c\n\x04slot\x18\x03 \x01(\x05\x12\x13\n\x0bvariable_id\x18\x04 \x01(\r\x12\r\n\x05value\x18\x05 \x01(\r\"\x0b\n\tStartGame\"^\n\x0ePlayerLeftGame\x12$\n\x0bplayer_uuid\x18\x01 \x01(\x0b\x32\x0f.stormgate.UUID\x12&\n\x06reason\x18\x02 \x01(\x0e\x32\x16.stormgate.LeaveReason\"u\n\x12\x43lientDisconnected\x12\x11\n\tclient_id\x18\x01 \x01(\x05\x12&\n\x06reason\x18\x02 \x01(\x0e\x32\x16.stormgate.LeaveReason\x12$\n\x0bplayer_uuid\x18\x03 \x01(\x0b\x32\x0f.stormgate.UUID\"Q\n\x10\x41ssignPlayerSlot\x12\x1d\n\x04uuid\x18\x01 \x01(\x0b\x32\x0f.stormgate.UUID\x12\x0c\n\x04slot\x18\x02 \x01(\x03\x12\x10\n\x08nickname\x18\x03 \x01(\t\"$\n\x04UUID\x12\r\n\x05part1\x18\x01 \x01(\x04\x12\r\n\x05part2\x18\x02 \x01(\x04*I\n\tMatchType\x12\x14\n\x10UnknownMatchType\x10\x00\x12\n\n\x06\x43ustom\x10\x01\x12\r\n\tRanked1v1\x10\x02\x12\x0b\n\x07\x43oop3vE\x10\x03*J\n\x0bLeaveReason\x12\x11\n\rUnknownReason\x10\x00\x12\r\n\tSurrender\x10\x01\x12\t\n\x05Leave\x10\x02\x12\x0e\n\nDisconnect\x10\x03\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'stormgate_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MATCHTYPE._serialized_start=1656
  _MATCHTYPE._serialized_end=1729
  _LEAVEREASON._serialized_start=1731
  _LEAVEREASON._serialized_end=1805
  _REPLAYCHUNK._serialized_start=31
  _REPLAYCHUNK._serialized_end=701
  _REPLAYCHUNK_WRAPPER._serialized_start=132
  _REPLAYCHUNK_WRAPPER._serialized_end=701
  _REPLAYCHUNK_WRAPPER_REPLAYCONTENT._serialized_start=207
  _REPLAYCHUNK_WRAPPER_REPLAYCONTENT._serialized_end=701
  _MAPDETAILS._serialized_start=703
  _MAPDETAILS._serialized_end=813
  _CLIENTCONNECTED._serialized_start=815
  _CLIENTCONNECTED._serialized_end=882
  _PLAYER._serialized_start=885
  _PLAYER._serialized_end=1023
  _PLAYER_PLAYERNAME._serialized_start=970
  _PLAYER_PLAYERNAME._serialized_end=1023
  _LOBBYCHANGESLOT._serialized_start=1026
  _LOBBYCHANGESLOT._serialized_end=1235
  _LOBBYCHANGESLOT_SLOTCHOICE._serialized_start=1101
  _LOBBYCHANGESLOT_SLOTCHOICE._serialized_end=1235
  _LOBBYCHANGESLOT_SLOTCHOICE_SPECIFICSLOT._serialized_start=1192
  _LOBBYCHANGESLOT_SLOTCHOICE_SPECIFICSLOT._serialized_end=1220
  _LOBBYSETVARIABLE._serialized_start=1237
  _LOBBYSETVARIABLE._serialized_end=1305
  _STARTGAME._serialized_start=1307
  _STARTGAME._serialized_end=1318
  _PLAYERLEFTGAME._serialized_start=1320
  _PLAYERLEFTGAME._serialized_end=1414
  _CLIENTDISCONNECTED._serialized_start=1416
  _CLIENTDISCONNECTED._serialized_end=1533
  _ASSIGNPLAYERSLOT._serialized_start=1535
  _ASSIGNPLAYERSLOT._serialized_end=1616
  _UUID._serialized_start=1618
  _UUID._serialized_end=1654
# @@protoc_insertion_point(module_scope)



"""shroudstone is a tool for managing Stormgate replays.

To get started renaming your replays, use [b]rename-replays --help[/b] to view
options or [b]rename-replays[/b] to jump straight in."""

import logging
import os
import platform
import subprocess
import sys
from pathlib import Path
from typing import Optional

import typer
from typing_extensions import Annotated

from shroudstone import __version__
from shroudstone.config import (
    DEFAULT_GENERIC_FORMAT,
    Config,
    DEFAULT_1v1_FORMAT,
    config_file,
)
from shroudstone.logging import configure_logging

app = typer.Typer(rich_markup_mode="rich", help=sys.modules[__name__].__doc__)

logger = logging.getLogger(__name__)


def version(value: bool):
    if value:
        typer.echo(f"Shroudstone v{__version__}")
        raise typer.Exit()


@app.callback()
def callback(
    ctx: typer.Context,
    version: Annotated[
        bool,
        typer.Option(
            "--version",
            help="Show version information for your shroudstone installation",
            callback=version,
        ),
    ] = False,
    debug: bool = False,
):
    configure_logging(debug=debug)


@app.command(rich_help_panel="Tools for nerds")
def get_replay_info(replay_file: typer.FileBinaryRead):
    """Extract information from a replay, outputting it in JSON format."""
    from shroudstone.replay import summarize_replay

    typer.echo(summarize_replay(replay_file).model_dump_json(indent=2))


@app.command(rich_help_panel="Tools for nerds")
def split_replay(replay_file: typer.FileBinaryRead, output_directory: Path):
    """Extract a stormgate replay into a directory containing individual protoscope messages."""
    from shroudstone.replay import split_replay

    output_directory.mkdir(exist_ok=True, parents=True)
    i = 0
    for i, chunk in enumerate(split_replay(replay_file)):
        (output_directory / f"{i:07d}.binpb").write_bytes(chunk)
    typer.echo(
        f"Wrote {i+1} replay messages in protoscope wire format to {output_directory}/."
    )


@app.command(rich_help_panel="Tools for nerds")
def dump_replay(replay_file: typer.FileBinaryRead):
    """Decode a replay and print a human-readable-ish representation of its contents."""
    from shroudstone.replay import split_replay
    from shroudstone.stormgate_pb2 import ReplayChunk

    for bytestring in split_replay(replay_file):
        chunk = ReplayChunk.FromString(bytestring)
        print(chunk.timestamp, chunk.client_id, chunk.inner.content)


@app.command(rich_help_panel="Tools for nerds")
def config_path():
    """Print the real path to the shroudstone configuration file."""
    if not config_file.exists():
        Config().save()
    typer.echo(config_file.resolve())


@app.command(rich_help_panel="Tools for nerds")
def edit_config(xdg_open: bool = False):
    """Open the shroudstone configuration file in your default text editor."""
    from shroudstone import renamer

    if not config_file.exists():
        logger.info("No config file found, doing our best to auto-generate one.")
        cfg = Config()
        # Prefill this path if possible:
        cfg.replay_dir = renamer.guess_replay_dir()
        cfg.save()

    realpath = config_file.resolve()
    if platform.system() == "Windows":
        # .resolve() is crucial when python is installed from the microsoft store
        subprocess.run(["cmd", "/c", f"start {realpath}"])
    else:
        if xdg_open:
            editor = "xdg-open"
        else:
            editor = os.environ.get("VISUAL", os.environ.get("EDITOR", "nano"))
        subprocess.run([editor, realpath])


@app.command(rich_help_panel="Replay renaming")
def gui():
    from shroudstone.gui import app

    app.run()


@app.command(rich_help_panel="Replay renaming")
def rename_replays(
    replay_dir: Annotated[
        Optional[Path],
        typer.Option(file_okay=False, dir_okay=True, exists=True, readable=True),
    ] = None,
    format_1v1: Annotated[
        Optional[str],
        typer.Option(
            help=f"Format string for 1v1 replays \n(e.g. '{DEFAULT_1v1_FORMAT}')"
        ),
    ] = None,
    format_generic: Annotated[
        Optional[str],
        typer.Option(
            help=f"Format string for other replays \n(e.g. '{DEFAULT_GENERIC_FORMAT}')"
        ),
    ] = None,
    backup: bool = True,
    dry_run: bool = False,
    reprocess: Annotated[
        bool, typer.Option(help="Reprocess old replays that have already been renamed")
    ] = False,
):
    """Automatically rename your replay files.

    The first time you run this, you will be asked for your replay directory if
    we can not find it automatically. This value is then stored in the
    configuration file for future runs.

    To customize the naming of your replays, you can provide a python format
    string in the --format option, or (preferably) edit the format string in
    your config file (see the [b]edit-config[/b] command).

    Format strings can use the following values:

    * us: Your nickname (as it appeared in the replay)
    * them: Opponent nickname (as it appeared in the replay)
    * f1: Faction/Race you played (Vanguard or Infernals or Maloc or Blockade)
    * f2: Faction/Race opponent played
    * time (datetime): Creation time of match
    * duration: Game duration (e.g. "15m10s")
    * result: Your game result (Win, Loss, Undecided)
    * map_name (str): Name of the map on which the game was played (extracted from replay file)
    * build_number (int): Build number of Stormgate version on which the game was played (extracted from replay file)
    """
    from shroudstone import renamer

    config = Config.load()
    if replay_dir is None:
        replay_dir = get_replay_dir(config)
    renamer.rename_replays(
        replay_dir=replay_dir,
        dry_run=dry_run,
        backup=backup,
        reprocess=reprocess,
        format_1v1=format_1v1 or config.replay_name_format_1v1,
        format_generic=format_generic or config.replay_name_format_generic,
    )


def get_replay_dir(config: Config) -> Path:
    from shroudstone.renamer import guess_replay_dir

    if config.replay_dir is None:
        config.replay_dir = guess_replay_dir()
        if config.replay_dir is None:
            typer.echo("Couldn't automatically determine your replay directory!")
            typer.echo(
                "It normally lives in your user directory under "
                r"AppData\Local\Stormgate\Saved\Replays."
            )
            typer.echo("If you know the correct path, please enter it now.")
            config.replay_dir = Path(typer.prompt("Path to replay directory"))
        config.save()
    return config.replay_dir



"""Main entrypoint for shroudstone.

By default, launches the GUI; unless any CLI arguments are provided, in which case we fall back to CLI."""

import sys


def main():
    if len(sys.argv) > 1:
        import shroudstone.cli

        shroudstone.cli.app()
    else:
        import shroudstone.gui.app

        shroudstone.gui.app.run()


if __name__ == "__main__":
    main()



"""Persistent configuration stored in standard user data directories"""

import os
import platform
from pathlib import Path
from typing import Optional

import yaml
from pydantic import BaseModel, ConfigDict


def _platform_data_dir() -> Path:
    if platform.system() == "Windows":
        return Path(os.environ["LOCALAPPDATA"])
    else:
        xdg_data_home = os.environ.get("XDG_DATA_HOME")
        if xdg_data_home:
            return Path(xdg_data_home)
        return Path.home() / ".local" / "share"


data_dir = _platform_data_dir() / "shroudstone"
data_dir.mkdir(parents=True, exist_ok=True)
config_file = data_dir / "config.yaml"

DEFAULT_1v1_FORMAT = "{time:%Y-%m-%d %H.%M} {result:.1} {duration} {us} {f1:.1}v{f2:.1} {them} - {map_name}.SGReplay"
DEFAULT_GENERIC_FORMAT = (
    "{time:%Y-%m-%d %H.%M} {duration} {players_with_factions} - {map_name}.SGReplay"
)
"""Default format string for new 1v1 replay filenames"""


class Config(BaseModel):
    replay_dir: Optional[Path] = None
    replay_name_format_1v1: str = DEFAULT_1v1_FORMAT
    replay_name_format_generic: str = DEFAULT_GENERIC_FORMAT
    minimize_to_tray: bool = False
    show_log_on_autorename: bool = False

    @staticmethod
    def load():
        if config_file.exists():
            with config_file.open("rt", encoding="utf-8") as f:
                content = yaml.load(f, Loader=yaml.SafeLoader)
                # Migrate old configs:
                if "replay_name_format" in content:
                    content["replay_name_format_1v1"] = content["replay_name_format"]
                return Config.model_validate(content)
        else:
            config = Config()
            config.save()
            return config

    def save(self):
        with config_file.open("wt", encoding="utf-8") as f:
            yaml.dump(self.model_dump(mode="json"), f, width=float("inf"))

    model_config = ConfigDict(extra="ignore")



import dataclasses
import logging
import platform
import tkinter as tk
from functools import partial
from pathlib import Path
from threading import Event, Thread, get_ident
from tkinter import ttk
from tkinter.filedialog import askdirectory
from tkinter.messagebox import askyesno, showinfo, showwarning
from typing import Optional

from PIL import Image
from pystray import Icon, Menu, MenuItem
from pystray._base import Icon as BaseIcon

from shroudstone import __version__, config, renamer
from shroudstone.gui.fonts import setup_style
from shroudstone.gui.logview import LogView
from shroudstone.logging import configure_logging

from .jobs import TkWithJobs

logger = logging.getLogger(__name__)
assets_dir = Path(__file__).parent / "assets"


class StringVar(tk.StringVar):
    def on_change(self, func):
        self.trace_add("write", func)


class BoolVar(tk.BooleanVar):
    def on_change(self, func):
        self.trace_add("write", func)


def field(factory, **kw):
    return dataclasses.field(default_factory=partial(factory, **kw))


@dataclasses.dataclass
class AppState:
    replay_dir: StringVar = field(StringVar)
    replay_name_format_1v1: StringVar = field(StringVar)
    replay_name_format_generic: StringVar = field(StringVar)
    reprocess: BoolVar = field(BoolVar)
    dry_run: BoolVar = field(BoolVar)
    autorename: BoolVar = field(BoolVar)
    minimize_to_tray: BoolVar = field(BoolVar)
    show_log_on_autorename: BoolVar = field(BoolVar)


class App(TkWithJobs):
    systray_icon: Optional[BaseIcon] = None
    systray_thread: Optional[Thread] = None
    tray_quit_event: Event
    vars: AppState

    def quit_app(self):
        logger.debug(f"Running quit_app in {get_ident()}")
        if self.systray_thread:
            logger.debug("Joining tray icon thread")
            self.systray_thread.join()
        logger.debug("Destroying main Tk app")
        self.destroy()

    def on_window_close(self):
        if self.systray_icon and self.vars.minimize_to_tray.get():
            self.withdraw()
        else:
            self.request_exit()

    def request_exit(self):
        if (not self.vars.autorename.get()) or askyesno(
            title="Shroudstone - Exit?",
            message="You have autorenaming enabled, which will stop functioning if you exit the program.\n"
            "Consider enabling 'Minimize to tray' if you want to keep auto-renaming without this window open.\n\n"
            "Are you sure you want to exit?\n",
        ):
            # Ask the tray icon to quit
            if self.systray_icon is not None:
                self.systray_icon.stop()
                # When it's done cleaning up, tray_quit_event will be set,
                # so no further action required
            else:
                self.quit_app()

    def setup_tray_icon(self):
        self.systray_thread = Thread(target=self._tray_icon_thread)
        self.systray_thread.start()

    def _tray_icon_thread(self):
        state = self.vars

        def toggle_autorename():
            state.autorename.set(not state.autorename.get())

        def show():
            self.deiconify()
            self.wm_state("normal")
            self.tkraise()

        def hide():
            self.withdraw()

        image = Image.open(assets_dir / "shroudstone.png")
        menu = Menu(
            MenuItem("Show", show, default=True),
            MenuItem("Minimize to tray", hide),
            MenuItem(
                "Auto-renaming",
                toggle_autorename,
                checked=lambda item: state.autorename.get(),
            ),
            MenuItem("Quit", lambda: icon.stop()),
        )
        icon = self.systray_icon = Icon(
            name="Shroudstone",
            icon=image,
            title="Shroudstone: Stormgate Replay Renamer",
            menu=menu,
        )

        @state.autorename.on_change
        def _(*_):
            icon.update_menu()

        logger.debug(f"icon.run() in thread {get_ident()}")
        icon.run()
        logger.debug("icon.run done")
        self.tray_quit_event.set()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.vars = AppState()
        self.protocol("WM_DELETE_WINDOW", self.on_window_close)
        logger.debug(f"Main thread is {get_ident()}")

        self.after(0, self.custom_startup)

    def custom_startup(self):
        """This method performs custom startup actions *after* the mainloop has started."""

        # The systray icon passes this message back from its thread when it's done:
        self.tray_quit_event = Event()

        def _check_quit_event():
            if self.tray_quit_event.is_set():
                self.quit_app()
            self.after(10, _check_quit_event)

        _check_quit_event()

        self.setup_tray_icon()


# TODO: Move the following methods inside our custom App class?
# We currently have some root.after calls in here, and some in our
# custom_startup method, so it's a bit of a mess. Definitely some kind of
# tidyup needs to be done.
def run():
    configure_logging()
    renamer.migrate()
    logger.info(
        "Keep this console open - it will show progress information during renaming."
    )
    cfg: config.Config = config.Config.load()

    root = App(className="Shroudstone")
    setup_style(root)
    setup_window_icon(root)

    if cfg.replay_dir is None:
        root.after(0, lambda: configure_replay_dir(root, root.vars, cfg))

    create_main_ui(root, cfg)
    root.mainloop()


def setup_window_icon(root: tk.Tk):
    if platform.system() == "Windows":
        # TODO: This .ico currently only has a 64x64px image in it, which looks
        # garbage when resized down to fit in window titlebars etc.
        window_icon = assets_dir / "shroudstone.ico"
        root.iconbitmap(str(window_icon))
    else:
        root.iconphoto(True, tk.PhotoImage(file=str(assets_dir / "shroudstone.png")))


def configure_replay_dir(root: TkWithJobs, state: AppState, cfg: config.Config):
    root.withdraw()  # Hide uninitialized main window while we show initial popups
    cfg.replay_dir = renamer.guess_replay_dir()
    if cfg.replay_dir is None:
        state.replay_dir.set("")
        showwarning(
            title="Stormgate Replay Directory",
            message="Could not automatically find your replay directory - "
            "please configure it manually on the next screen.",
        )
    else:
        state.replay_dir.set(str(cfg.replay_dir))
        showinfo(
            title="Stormgate Replay Directory",
            message=f"Detected your replay directory as {cfg.replay_dir}."
            " If this is incorrect, please configure it manually on the next screen.",
        )
    cfg.save()
    root.deiconify()


def rename_replays_wrapper(*args, **kwargs):
    try:
        return renamer.rename_replays(*args, **kwargs)
    except Exception:
        logger.exception(
            "Unexpected error occurred! Please report here: "
            "https://github.com/acarapetis/shroudstone/issues"
        )


def create_main_ui(root: App, cfg: config.Config):
    state = root.vars

    def reload_config():
        nonlocal cfg
        cfg = config.Config.load()
        state.replay_dir.set(str(cfg.replay_dir or ""))
        state.replay_name_format_1v1.set(cfg.replay_name_format_1v1)
        state.replay_name_format_generic.set(cfg.replay_name_format_generic)
        state.minimize_to_tray.set(cfg.minimize_to_tray)
        state.show_log_on_autorename.set(cfg.show_log_on_autorename)

    def save_config():
        cfg.save()

    root.title(f"Shroudstone v{__version__} - Stormgate Replay Renamer")

    config_frame = ttk.LabelFrame(root, text="Configuration")
    config_frame.pack(fill="x", padx=5, pady=5)

    form = ttk.Frame(config_frame)
    form.pack(fill="x")

    form.columnconfigure(0, weight=0)
    form.columnconfigure(1, weight=1)

    ttk.Label(form, text="Replay Directory", justify="right").grid(
        row=0, column=0, sticky="E", padx=2, pady=2
    )
    replay_dir_row = ttk.Frame(form)
    replay_dir_row.grid(row=0, column=1, sticky="NSEW")
    replay_dir_entry = ttk.Entry(
        replay_dir_row, width=50, textvariable=state.replay_dir
    )
    replay_dir_entry.pack(side="left", fill="both", ipadx=2, ipady=2, expand=True)

    replay_dir_error = ttk.Label(form)
    replay_dir_error.grid(row=1, column=1, sticky="WE", ipadx=5, ipady=5)

    def browse_replay_dir():
        current = Path(state.replay_dir.get())
        initial = current if current.exists() else None
        new = askdirectory(
            title="Stormgate Replay Directory", initialdir=initial, mustexist=True
        )
        state.replay_dir.set(new)

    def guess_replay_dir():
        rd = renamer.guess_replay_dir()
        if rd is not None:
            state.replay_dir.set(str(rd))

    ttk.Button(replay_dir_row, text="Browse", command=browse_replay_dir).pack(
        side="left", fill="y", padx=2, ipadx=2, ipady=2
    )

    ttk.Button(replay_dir_row, text="Autodetect", command=guess_replay_dir).pack(
        side="left", fill="y", padx=2, ipadx=2, ipady=2
    )

    @state.replay_dir.on_change
    def _(*_):
        path = Path(state.replay_dir.get())
        if path.is_dir():
            replay_dir_error.configure(text="Looks good!", background="#66ff66")
            cfg.replay_dir = path
            save_config_button.configure(state="normal")
            rename_button.configure(state="normal")
        else:
            replay_dir_error.configure(
                text="Directory does not exist!", background="#ff6666"
            )
            save_config_button.configure(state="disabled")
            rename_button.configure(state="disabled")

    @state.replay_name_format_1v1.on_change
    def _(*_):
        fstr = state.replay_name_format_1v1.get()
        try:
            renamer.validate_format_string(fstr, type="1v1")
        except ValueError as e:
            format_error_1v1.configure(text=f"Error: {e}", background="#ff6666")
            save_config_button.configure(state="disabled")
            rename_button.configure(state="disabled")
        else:
            format_error_1v1.configure(text="Looks good!", background="#66ff66")
            cfg.replay_name_format_1v1 = fstr
            save_config_button.configure(state="normal")
            rename_button.configure(state="normal")

    @state.replay_name_format_generic.on_change
    def _(*_):
        fstr = state.replay_name_format_generic.get()
        try:
            renamer.validate_format_string(fstr, type="generic")
        except ValueError as e:
            format_error_generic.configure(text=f"Error: {e}", background="#ff6666")
            save_config_button.configure(state="disabled")
            rename_button.configure(state="disabled")
        else:
            format_error_generic.configure(text="Looks good!", background="#66ff66")
            cfg.replay_name_format_generic = fstr
            save_config_button.configure(state="normal")
            rename_button.configure(state="normal")

    ttk.Label(form, text="New Filename Format (1v1)", justify="right").grid(
        row=3, column=0, sticky="E", padx=2, pady=2
    )
    format_entry_1v1 = ttk.Entry(
        form,
        width=100,
        textvariable=state.replay_name_format_1v1,
    )
    format_entry_1v1.grid(row=3, column=1, sticky="WE", ipadx=2, ipady=2)
    format_error_1v1 = ttk.Label(form)
    format_error_1v1.grid(row=4, column=1, sticky="WE", ipadx=5, ipady=5)

    ttk.Label(form, text="New Filename Format (other)", justify="right").grid(
        row=5, column=0, sticky="E", padx=2, pady=2
    )
    format_entry_generic = ttk.Entry(
        form,
        width=100,
        textvariable=state.replay_name_format_generic,
    )
    format_entry_generic.grid(row=5, column=1, sticky="WE", ipadx=2, ipady=2)
    format_error_generic = ttk.Label(form)
    format_error_generic.grid(row=6, column=1, sticky="WE", ipadx=5, ipady=5)

    config_buttons = ttk.Frame(config_frame)
    config_buttons.pack(fill="x")

    save_config_button = ttk.Button(
        config_buttons, text="Save Config", command=save_config
    )
    save_config_button.pack(side="right", fill="both", padx=3, pady=3)

    load_config = ttk.Button(
        config_buttons, text="Reload Config", command=reload_config
    )
    load_config.pack(side="right", fill="both", padx=3, pady=3)

    renaming = False

    def rename_replays(auto: bool = False):
        nonlocal renaming
        if renaming:
            return
        renaming = True
        show_log = state.show_log_on_autorename.get() or not auto
        btn_text = "Renaming in progress"
        if show_log:
            btn_text += " - See log window"
            log_view.deiconify()
        rename_button.config(text=btn_text, state="disabled")

        def callback(_):
            rename_button.config(
                text="Rename My Replays Now",
                state="normal",
            )
            nonlocal renaming
            renaming = False

        cfg.replay_dir = Path(state.replay_dir.get())
        cfg.replay_name_format_1v1 = state.replay_name_format_1v1.get()
        cfg.replay_name_format_generic = state.replay_name_format_generic.get()

        root.jobs.submit(
            rename_replays_wrapper,
            callback,
            replay_dir=cfg.replay_dir,
            format_1v1=cfg.replay_name_format_1v1,
            format_generic=cfg.replay_name_format_generic,
            reprocess=state.reprocess.get(),
            dry_run=state.dry_run.get(),
        )

    options_frame = ttk.LabelFrame(root, text="Options")
    options_frame.pack(fill="x", padx=5, pady=5)

    reprocess_cb = tk.Checkbutton(
        options_frame,
        variable=state.reprocess,
        text="Reprocess replays that have already been renamed",
    )
    reprocess_cb.pack(anchor="w")
    dry_run_cb = tk.Checkbutton(
        options_frame,
        variable=state.dry_run,
        text="Dry run - don't actually rename file, just show output",
    )
    dry_run_cb.pack(anchor="w")

    rename_button = ttk.Button(
        root,
        text="Rename My Replays Now",
        command=rename_replays,
    )

    rename_button.pack(fill="x")

    gui_toggles = ttk.Frame(root)
    gui_toggles.pack(fill="x")

    ttk.Checkbutton(
        gui_toggles, text="Automatically rename new replays", variable=state.autorename
    ).pack(side="left", padx=5, pady=5)

    autorename_ref = None

    @state.autorename.on_change
    def _(*_):
        nonlocal autorename_ref
        if autorename_ref is not None:
            root.after_cancel(autorename_ref)
            autorename_ref = None
        if state.autorename.get():

            def doit():
                nonlocal autorename_ref
                rename_replays(auto=True)
                autorename_ref = root.after(30000, doit)

            doit()

    ttk.Checkbutton(
        gui_toggles, text="Minimize to tray on close", variable=state.minimize_to_tray
    ).pack(side="left", padx=5, pady=5)

    ttk.Checkbutton(
        gui_toggles,
        text="Show Log when auto-renaming",
        variable=state.show_log_on_autorename,
    ).pack(side="left", padx=5, pady=5)

    ttk.Button(gui_toggles, text="Exit", command=root.request_exit).pack(
        side="right", padx=5, pady=5
    )

    @state.replay_dir.on_change
    def _(*_):
        cfg.replay_dir = Path(state.replay_dir.get())

    @state.minimize_to_tray.on_change
    def _(*_):
        cfg.minimize_to_tray = state.minimize_to_tray.get()

    reload_config()

    log_view = LogView(root)
    log_view.withdraw()
    log_view.protocol("WM_DELETE_WINDOW", log_view.clear_and_close)



import logging
import tkinter as tk
from queue import Empty, Queue
from tkinter.font import Font
from tkinter.scrolledtext import ScrolledText

from .fonts import first_available_font


# We pipe log messages through a queue so that only the main thread touches Tk
class QueueHandler(logging.Handler):
    queue: Queue

    def __init__(self, queue: Queue, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.queue = queue

    def emit(self, record: logging.LogRecord):
        self.queue.put(record)


class LogView(tk.Toplevel):
    textbox: ScrolledText
    queue: Queue[logging.LogRecord]
    handler: QueueHandler

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        mono = first_available_font(
            "Iosevka",
            "DejaVu Sans Mono",
            "Ubuntu Mono",
            "Monaco",
            "Consolas",
            "Monospace",
        )
        self.title("Shroudstone Log")
        self.geometry("1024x600")
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.textbox = ScrolledText(
            self,
            font=Font(name="Term", family=mono, size=12),
            background="black",
            foreground="white",
        )
        self.textbox.grid(column=0, row=0, sticky="NSEW")
        self.textbox.configure(state="disabled")
        self.textbox.tag_config(
            "error",
            foreground="red",
            font=Font(name="TermBold", family=mono, size=12, weight="bold"),
        )
        self.textbox.tag_config("warning", foreground="orange")
        self.queue = Queue()
        self.handler = QueueHandler(self.queue)
        logging.getLogger().addHandler(self.handler)
        self.after(0, self.tick)

    def destroy(self):
        logging.getLogger().removeHandler(self.handler)
        return super().destroy()

    def tick(self):
        self.textbox.configure(state="normal")
        while True:
            try:
                record = self.queue.get_nowait()
            except Empty:
                break
            else:
                self.handle_record(record)
        self.textbox.configure(state="disabled")
        self.after(10, self.tick)

    def handle_record(self, record: logging.LogRecord):
        tags = []
        if record.levelno >= logging.WARNING:
            tags = ["warning"]
        if record.levelno >= logging.ERROR:
            tags = ["error"]
        self.textbox.insert(tk.END, f"{record.msg}\n", *tags)
        # if record.exc_info:
        #     tb = "\n".join(traceback.format_tb(record.exc_info[2]))
        #     self.textbox.insert(tk.END, tb, *tags)
        if record.exc_text:
            self.textbox.insert(tk.END, record.exc_text + "\n", *tags)

    def clear_and_close(self):
        self.textbox.configure(state="normal")
        self.textbox.delete("1.0", tk.END)
        self.textbox.configure(state="disabled")
        self.withdraw()



"""Simple framework for running asynchronous jobs in a Tk app in threads."""

from __future__ import annotations

import tkinter as tk
from concurrent.futures import Future, ThreadPoolExecutor
from dataclasses import dataclass
from typing import Callable, List

from typing_extensions import Generic, ParamSpec, TypeVar


class TkWithJobs(tk.Tk):
    jobs: JobManager

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.jobs = JobManager(self)

    def debounce(self, timeout: int = 500):
        """Use this Tk's event loop to debounce a 0-arg function.

        Useful for preventing on-change/on-keypress event handlers from firing too often."""
        state: dict = {"timer": None}

        def decorator(func):
            def clear_timer_and_run(*_):
                state["timer"] = None
                func()

            def start_timer(*_):
                if state["timer"] is not None:
                    self.after_cancel(state["timer"])
                state["timer"] = self.after(timeout, clear_timer_and_run)

            return start_timer

        return decorator


T = TypeVar("T")
P = ParamSpec("P")


@dataclass(frozen=True)
class FutureContinuation(Generic[T]):
    """A future (representing a job running in a thread pool) along with a
    callback that will be executed in the main thread when the future
    completes. The callback should be passed a single argument: the result
    value of the future."""

    future: Future[T]
    continuation: Callable[[T], None]


class JobManager:
    """Integrates a threadpool with a Tk widget, checking the list of futures for
    completed results every 50ms."""

    jobs: List[FutureContinuation]
    pool: ThreadPoolExecutor
    root: tk.Tk
    _timer_id: str

    def __init__(self, tk):
        self.jobs = []
        self.pool = ThreadPoolExecutor()
        self.root = tk
        self.tick()

    def tick(self):
        remaining: List[FutureContinuation] = []
        ready: List[FutureContinuation] = []
        for f in self.jobs:
            if f.future.done():
                ready.append(f)
            else:
                remaining.append(f)
        self.jobs = remaining
        for f in ready:
            f.continuation(f.future.result())
        self._timer_id = self.root.after(50, self.tick)

    def submit(
        self,
        func: Callable[P, T],
        callback: Callable[[T], None],
        *args: P.args,
        **kwargs: P.kwargs,
    ):
        future = self.pool.submit(func, *args, **kwargs)
        self.jobs.append(FutureContinuation(future, callback))

    def destroy(self):
        """Call this before destroying self.tk to avoid broken reference
        errors."""
        self.root.after_cancel(self._timer_id)
